# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib
import shutil

import pytest
from rocrate.utils import get_norm_value


THIS_DIR = pathlib.Path(__file__).absolute().parent
TEST_DATA_NAME = 'test-data'
BASE_URL = 'https://w3id.org/ro/crate'
VERSION = '1.1'
LEGACY_VERSION = '1.0'


class Helpers:

    PROFILE = f"{BASE_URL}/{VERSION}"
    LEGACY_PROFILE = f"{BASE_URL}/{LEGACY_VERSION}"
    WORKFLOW_PROFILE = "https://w3id.org/workflowhub/workflow-ro-crate/1.0"
    METADATA_FILE_NAME = 'ro-crate-metadata.json'
    LEGACY_METADATA_FILE_NAME = 'ro-crate-metadata.jsonld'
    WORKFLOW_TYPES = {"File", "SoftwareSourceCode", "ComputationalWorkflow"}
    WORKFLOW_DESC_TYPES = {"File", "SoftwareSourceCode", "HowTo"}
    LEGACY_WORKFLOW_TYPES = {"File", "SoftwareSourceCode", "Workflow"}
    PREVIEW_FILE_NAME = "ro-crate-preview.html"

    @classmethod
    def read_json_entities(cls, crate_base_path):
        metadata_path = pathlib.Path(crate_base_path) / cls.METADATA_FILE_NAME
        with open(metadata_path, "rt") as f:
            json_data = json.load(f)
        return {_["@id"]: _ for _ in json_data["@graph"]}

    @classmethod
    def check_crate(cls, json_entities, root_id="./", data_entity_ids=None):
        assert root_id in json_entities
        root = json_entities[root_id]
        assert root["@type"] == "Dataset"
        assert cls.METADATA_FILE_NAME in json_entities
        metadata = json_entities[cls.METADATA_FILE_NAME]
        assert metadata["@type"] == "CreativeWork"
        assert cls.PROFILE in get_norm_value(metadata, "conformsTo")
        assert metadata["about"] == {"@id": root_id}
        if data_entity_ids:
            data_entity_ids = set(data_entity_ids)
            assert data_entity_ids.issubset(json_entities)
            assert "hasPart" in root
            assert data_entity_ids.issubset([_["@id"] for _ in root["hasPart"]])

    @classmethod
    def check_wf_crate(cls, json_entities, wf_file_name, root_id="./"):
        cls.check_crate(json_entities, root_id=root_id)
        assert json_entities[root_id]["mainEntity"]["@id"] == wf_file_name
        assert wf_file_name in json_entities
        wf_entity = json_entities[wf_file_name]
        assert isinstance(wf_entity["@type"], list)
        assert cls.WORKFLOW_TYPES.issubset(wf_entity["@type"])
        assert "programmingLanguage" in wf_entity
        metadata = json_entities[cls.METADATA_FILE_NAME]
        assert cls.WORKFLOW_PROFILE in get_norm_value(metadata, "conformsTo")


@pytest.fixture
def helpers():
    return Helpers


# pytest's default tmpdir returns a py.path object
@pytest.fixture
def tmpdir(tmpdir):
    return pathlib.Path(tmpdir)


@pytest.fixture
def test_data_dir(tmpdir):
    d = tmpdir / TEST_DATA_NAME
    shutil.copytree(THIS_DIR / TEST_DATA_NAME, d)
    return d



#!/usr/bin/env python

# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from codecs import open
from os import path
import re

# https://www.python.org/dev/peps/pep-0440/#appendix-b-parsing-version-strings-with-regular-expressions  # noqa
PEP440_PATTERN = r"([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?"  # noqa


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    required = f.read().splitlines()

with open(path.join(here, 'rocrate', '_version.py'), encoding='utf-8') as f:
    # "parse" rocrate/_version.py which MUST have this pattern
    # __version__ = "0.1.1"
    # see https://www.python.org/dev/peps/pep-0440
    v = f.read().strip()
    m = re.match(r'^__version__ = "(' + PEP440_PATTERN + ')"$', v)
    if not m:
        msg = ('rocrate/_version.py did not match pattern '
               '__version__ = "0.1.2"  (see PEP440):\n') + v
        raise Exception(msg)
    __version__ = m.group(1)


setup(
    name='rocrate',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    version=__version__,  # update in rocrate/_version.py
    description='RO-Crate metadata generator/parser',
    long_description_content_type='text/markdown',
    long_description=long_description,
    author=", ".join((
        'Paul De Geest',
        'Bert Droesbeke',
        'Ignacio Eguinoa',
        'Alban Gaignard',
        'Sebastiaan Huber',
        'Bruno Kinoshita',
        'Simone Leo',
        'Luca Pireddu',
        'Laura Rodríguez-Navas',
        'Raül Sirvent',
        'Stian Soiland-Reyes'
    )),
    python_requires='>=3.7',
    author_email='stain@apache.org',
    package_data={'': ['data/*.jsonld', 'templates/*.j2']},
    # SPDX, pending https://github.com/pombredanne/spdx-pypi-pep/pull/2
    license="Apache-2.0",
    url='https://github.com/ResearchObject/ro-crate-py/',
    download_url=('https://github.com/researchobject/ro-crate-py/archive/'
                  f'{__version__}.tar.gz'),
    keywords="researchobject ro-crate ro metadata jsonld",
    install_requires=[required],
    classifiers=[
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: System :: Archiving',
        'Topic :: System :: Archiving :: Packaging',
    ],
    entry_points={
        "console_scripts": ["rocrate=rocrate.cli:cli"],
    },
)



# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .contextentity import ContextEntity


class ComputerLanguage(ContextEntity):

    def _empty(self):
        return {
            "@id": self.id,
            "@type": 'ComputerLanguage'
        }

    @property
    def name(self):
        return self.get("name")

    @name.setter
    def name(self, name):
        self["name"] = name

    @property
    def alternateName(self):
        return self.get("alternateName")

    @alternateName.setter
    def alternateName(self, alternateName):
        self["alternateName"] = alternateName

    @property
    def identifier(self):
        return self.get("identifier")

    @identifier.setter
    def identifier(self, identifier):
        self["identifier"] = identifier

    @property
    def url(self):
        return self.get("url")

    @url.setter
    def url(self, url):
        self["url"] = url

    # Not listed as a property in "https://schema.org/ComputerLanguage"
    @property
    def version(self):
        return self.get("version")

    @version.setter
    def version(self, version):
        self["version"] = version


# See https://w3id.org/workflowhub/workflow-ro-crate/1.0
# (note that it does not specify "version")


def cwl(crate, version=None):
    id_ = "https://w3id.org/workflowhub/workflow-ro-crate#cwl"
    identifier = "https://w3id.org/cwl/"
    if version:
        identifier = f"{identifier}v{version}/"
    properties = {
        "name": "Common Workflow Language",
        "alternateName": "CWL",
        "identifier": {
            "@id": identifier
        },
        "url": {
            "@id": "https://www.commonwl.org/"
        },
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier=id_, properties=properties)


def galaxy(crate, version=None):
    id_ = "https://w3id.org/workflowhub/workflow-ro-crate#galaxy"
    properties = {
        "name": "Galaxy",
        "identifier": {
            "@id": "https://galaxyproject.org/"
        },
        "url": {
            "@id": "https://galaxyproject.org/"
        }
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier=id_, properties=properties)


def knime(crate, version=None):
    id_ = "https://w3id.org/workflowhub/workflow-ro-crate#knime"
    properties = {
        "name": "KNIME",
        "identifier": {
            "@id": "https://www.knime.com/"
        },
        "url": {
            "@id": "https://www.knime.com/"
        }
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier=id_, properties=properties)


def nextflow(crate, version=None):
    id_ = "https://w3id.org/workflowhub/workflow-ro-crate#nextflow"
    properties = {
        "name": "Nextflow",
        "identifier": {
            "@id": "https://www.nextflow.io/"
        },
        "url": {
            "@id": "https://www.nextflow.io/"
        }
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier=id_, properties=properties)


def snakemake(crate, version=None):
    id_ = "https://w3id.org/workflowhub/workflow-ro-crate#snakemake"
    properties = {
        "name": "Snakemake",
        "identifier": {
            "@id": "https://doi.org/10.1093/bioinformatics/bts480"
        },
        "url": {
            "@id": "https://snakemake.readthedocs.io"
        }
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier=id_, properties=properties)


def compss(crate, version=None):
    properties = {
        "name": "COMPSs Programming Model",
        "alternateName": "COMPSs",
        "url": "http://compss.bsc.es/",
        "citation": "https://doi.org/10.1007/s10723-013-9272-5"
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier="#compss", properties=properties)


def autosubmit(crate, version=None):
    properties = {
        "name": "Autosubmit",
        "alternateName": "AS",
        "url": "https://autosubmit.readthedocs.io/",
        "citation": "https://doi.org/10.1109/HPCSim.2016.7568429"
    }
    if version:
        properties["version"] = version
    return ComputerLanguage(crate, identifier="#autosubmit", properties=properties)


LANG_MAP = {
    "cwl": cwl,
    "galaxy": galaxy,
    "knime": knime,
    "nextflow": nextflow,
    "snakemake": snakemake,
    "compss": compss,
    "autosubmit": autosubmit,
}


def get_lang(crate, name, version=None):
    try:
        func = LANG_MAP[name.lower()]
    except KeyError:
        raise ValueError(f"Unknown language: {name}")
    return func(crate, version=version)



# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from copy import deepcopy

from rocrate.metadata import find_root_entity_id


@pytest.mark.parametrize("root,basename", [
    ("", "ro-crate-metadata.json"),
    ("", "ro-crate-metadata.jsonld"),
    ("https://example.org/crate/", "ro-crate-metadata.json"),
    ("https://example.org/crate/", "ro-crate-metadata.jsonld"),
    ("", "bad-name.json"),
])
def test_find_root(root, basename):
    metadata_id = root + basename
    root_id = root or "./"
    entities = {_["@id"]: _ for _ in [
        {
            "@id": metadata_id,
            "@type": "CreativeWork",
            "about": {"@id": root_id},
            "conformsTo": [
                {"@id": "https://w3id.org/ro/crate/1.1"},
                {"@id": "https://example.org/fancy-ro-crate/1.0"},
            ]
        },
        {
            "@id": root_id,
            "@type": "Dataset",
        },
    ]}
    if basename not in {"ro-crate-metadata.json", "ro-crate-metadata.jsonld"}:
        with pytest.raises(KeyError):
            find_root_entity_id(entities)
    else:
        assert find_root_entity_id(entities) == (metadata_id, root_id)


def test_find_root_bad_entities():
    orig_entities = {
        "ro-crate-metadata.json": {
            "@id": "ro-crate-metadata.json",
            "@type": "CreativeWork",
            "about": {"@id": "./"},
            "conformsTo": {"@id": "https://w3id.org/ro/crate/1.1"},
        },
        "./": {
            "@id": "./",
            "@type": "Dataset",
        },
    }
    # missing "about"
    entities = deepcopy(orig_entities)
    del entities["ro-crate-metadata.json"]["about"]
    with pytest.raises(ValueError, match="does not reference"):
        find_root_entity_id(entities)
    # "about" does not reference the root entity
    entities = deepcopy(orig_entities)
    for about in "http://example.org", {"@id": "http://example.org"}:
        entities["ro-crate-metadata.json"]["about"] = about
        with pytest.raises(ValueError, match="does not reference"):
            find_root_entity_id(entities)
    # metadata type is not CreativeWork
    entities = deepcopy(orig_entities)
    entities["ro-crate-metadata.json"]["@type"] = "Thing"
    with pytest.raises(ValueError, match="must be of type"):
        find_root_entity_id(entities)
    # root type is not Dataset
    entities = deepcopy(orig_entities)
    entities["./"]["@type"] = "Thing"
    with pytest.raises(ValueError, match="must have"):
        find_root_entity_id(entities)


@pytest.mark.filterwarnings("ignore")
def test_find_root_multiple_entries():
    orig_entities = {
        "http://example.org/ro-crate-metadata.json": {
            "@id": "http://example.org/ro-crate-metadata.json",
            "@type": "CreativeWork",
            "about": {"@id": "http://example.org/"},
            "conformsTo": {"@id": "https://w3id.org/ro/crate/1.1"},
        },
        "http://example.org/": {
            "@id": "http://example.org/",
            "@type": "Dataset",
            "hasPart": [
                {"@id": "http://example.com/"},
                {"@id": "http://example.com/ro-crate-metadata.json"}
            ]
        },
        "http://example.com/ro-crate-metadata.json": {
            "@id": "http://example.com/ro-crate-metadata.json",
            "@type": "CreativeWork",
            "about": {"@id": "http://example.com/"},
            "conformsTo": {"@id": "https://w3id.com/ro/crate/1.1"},
        },
        "http://example.com/": {
            "@id": "http://example.com/",
            "@type": "Dataset",
        },
    }

    def check_finds_org(entities):
        m_id, r_id = find_root_entity_id(entities)
        assert m_id == "http://example.org/ro-crate-metadata.json"
        assert r_id == "http://example.org/"

    def check_picks_one(entities):
        m_id, r_id = find_root_entity_id(entities)
        assert m_id in [f"http://example.{_}/ro-crate-metadata.json" for _ in ("org", "com")]
        assert r_id in [f"http://example.{_}/" for _ in ("org", "com")]

    check_finds_org(orig_entities)
    # no root candidate contains the other one
    mod_entities = deepcopy(orig_entities)
    del mod_entities["http://example.org/"]["hasPart"]
    check_picks_one(mod_entities)
    # each root candidate contains the other one
    mod_entities = deepcopy(orig_entities)
    mod_entities["http://example.com/"]["hasPart"] = [
        {"@id": "http://example.org/"},
        {"@id": "http://example.org/ro-crate-metadata.json"}
    ]
    check_picks_one(mod_entities)
    # "about" does not reference the root entity
    mod_entities = deepcopy(orig_entities)
    for about in "http://google.com", {"@id": "http://google.com"}:
        mod_entities["http://example.com/ro-crate-metadata.json"]["about"] = about
        check_finds_org(mod_entities)
    # metadata type is not CreativeWork
    mod_entities = deepcopy(orig_entities)
    mod_entities["http://example.com/ro-crate-metadata.json"]["@type"] = "Thing"
    check_finds_org(mod_entities)
    # root type is not Dataset
    mod_entities = deepcopy(orig_entities)
    mod_entities["http://example.com/"]["@type"] = "Thing"
    check_finds_org(mod_entities)


def test_find_root_multiple_types():
    entities = {_["@id"]: _ for _ in [
        {
            "@id": "ro-crate-metadata.json",
            "@type": "CreativeWork",
            "about": {"@id": "./"},
            "conformsTo": {"@id": "https://w3id.org/ro/crate/1.1"},
        },
        {
            "@id": "./",
            "@type": ["Dataset", "RepositoryCollection"],
        },
    ]}
    m_id, r_id = find_root_entity_id(entities)
    assert m_id == "ro-crate-metadata.json"
    assert r_id == "./"
    # "Dataset" not included
    del entities["./"]["@type"][0]
    with pytest.raises(ValueError):
        find_root_entity_id(entities)
    # Check we're not trying to be too clever
    entities["./"]["@type"] = "NotADataset"
    with pytest.raises(ValueError):
        find_root_entity_id(entities)



# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from rocrate.rocrate import ROCrate, make_workflow_rocrate

WF_CRATE = "https://w3id.org/workflowhub/workflow-ro-crate"


def test_galaxy_wf_crate(test_data_dir, tmpdir, helpers):
    wf_id = 'test_galaxy_wf.ga'
    wf_path = test_data_dir / wf_id
    wf_crate = make_workflow_rocrate(wf_path, wf_type='Galaxy')
    assert isinstance(wf_crate, ROCrate)

    wf = wf_crate.dereference(wf_id)
    assert wf._default_type == "ComputationalWorkflow"
    assert wf_crate.mainEntity is wf
    lang = wf_crate.dereference(f"{WF_CRATE}#galaxy")
    assert hasattr(lang, "name")
    assert "version" not in lang
    assert wf.get("programmingLanguage") is lang
    assert wf.get("subjectOf") is not None
    assert helpers.WORKFLOW_DESC_TYPES.issubset(wf["subjectOf"].type)

    out_path = tmpdir / 'ro_crate_out'
    out_path.mkdir()
    wf_crate.write(out_path)
    json_entities = helpers.read_json_entities(out_path)
    helpers.check_wf_crate(json_entities, wf_id)
    wf_entity = json_entities[wf_id]
    assert "subjectOf" in wf_entity
    abstract_wf_id = wf_entity["subjectOf"]["@id"]
    abstract_wf_entity = json_entities[abstract_wf_id]
    assert helpers.WORKFLOW_DESC_TYPES.issubset(abstract_wf_entity["@type"])

    wf_out_path = out_path / wf_id
    assert wf_out_path.exists()
    with open(wf_path) as f1, open(wf_out_path) as f2:
        assert f1.read() == f2.read()

    abstract_wf_out_path = out_path / abstract_wf_id
    assert abstract_wf_out_path.exists()


def test_cwl_wf_crate(test_data_dir, tmpdir, helpers):
    wf_id = 'sample_cwl_wf.cwl'
    wf_path = test_data_dir / wf_id
    wf_crate = make_workflow_rocrate(wf_path, wf_type='CWL')
    assert isinstance(wf_crate, ROCrate)

    wf = wf_crate.dereference(wf_id)
    assert wf_crate.mainEntity is wf
    lang = wf_crate.dereference(f"{WF_CRATE}#cwl")
    assert hasattr(lang, "name")
    assert "version" not in lang
    assert wf.get("programmingLanguage") is lang
    assert "subjectOf" not in wf

    out_path = tmpdir / 'ro_crate_out'
    out_path.mkdir()
    wf_crate.write(out_path)
    json_entities = helpers.read_json_entities(out_path)
    helpers.check_wf_crate(json_entities, wf_id)

    wf_out_path = out_path / wf_id
    assert wf_out_path.exists()
    with open(wf_path) as f1, open(wf_out_path) as f2:
        assert f1.read() == f2.read()


def test_create_wf_include(test_data_dir, tmpdir, helpers):
    wf_id = 'test_galaxy_wf.ga'
    wf_path = test_data_dir / wf_id
    extra_file1 = test_data_dir / 'test_file_galaxy.txt'
    extra_file2 = test_data_dir / 'test_file_galaxy2.txt'
    files_list = [extra_file1, extra_file2]
    wf_crate = make_workflow_rocrate(
        wf_path, wf_type='Galaxy', include_files=files_list
    )
    assert isinstance(wf_crate, ROCrate)

    wf = wf_crate.dereference(wf_id)
    assert wf_crate.mainEntity is wf
    lang = wf_crate.dereference(f"{WF_CRATE}#galaxy")
    assert hasattr(lang, "name")
    assert "version" not in lang
    assert wf.get("programmingLanguage") is lang
    assert wf.get("subjectOf") is not None
    assert helpers.WORKFLOW_DESC_TYPES.issubset(wf["subjectOf"].type)
    assert wf_crate.dereference(extra_file1.name) is not None
    assert wf_crate.dereference(extra_file2.name) is not None

    out_path = tmpdir / 'ro_crate_out'
    out_path.mkdir()
    wf_crate.write(out_path)
    json_entities = helpers.read_json_entities(out_path)
    helpers.check_wf_crate(json_entities, wf_id)

    wf_out_path = out_path / wf_id
    file1 = out_path / extra_file1.name
    file2 = out_path / extra_file2.name
    assert wf_out_path.exists()
    with open(wf_path) as f1, open(wf_out_path) as f2:
        assert f1.read() == f2.read()
    assert file1.exists()
    with open(extra_file1) as f1, open(file1) as f2:
        assert f1.read() == f2.read()
    assert file2.exists()
    with open(extra_file2) as f1, open(file2) as f2:
        assert f1.read() == f2.read()


@pytest.mark.parametrize("lang_version", [None, "1.2"])
def test_cwl_lang_version(test_data_dir, lang_version):
    wf_id = 'sample_cwl_wf.cwl'
    wf_path = test_data_dir / wf_id
    crate = ROCrate()
    workflow = crate.add_workflow(wf_path, wf_id, lang_version=lang_version)
    lang = workflow["programmingLanguage"]
    lang_id = lang["identifier"]
    if lang_version is None:
        assert lang_id == "https://w3id.org/cwl/"
        assert "version" not in lang
    else:
        assert lang_id == "https://w3id.org/cwl/v1.2/"
        assert lang["version"] == "1.2"



#!/usr/bin/env python

# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import shutil
import urllib.request
import warnings
from http.client import HTTPResponse
from io import BytesIO, StringIO

from .file_or_dir import FileOrDir
from ..utils import is_url, iso_now


class File(FileOrDir):

    def _empty(self):
        val = {
            "@id": self.id,
            "@type": 'File'
        }
        return val

    def write(self, base_path):
        out_file_path = Path(base_path) / self.id
        if isinstance(self.source, (BytesIO, StringIO)):
            out_file_path.parent.mkdir(parents=True, exist_ok=True)
            mode = 'w' + ('b' if isinstance(self.source, BytesIO) else 't')
            with open(out_file_path, mode) as out_file:
                out_file.write(self.source.getvalue())
        elif is_url(str(self.source)):
            if self.fetch_remote or self.validate_url:
                with urllib.request.urlopen(self.source) as response:
                    if self.validate_url:
                        if isinstance(response, HTTPResponse):
                            self._jsonld.update({
                                'contentSize': response.getheader('Content-Length'),
                                'encodingFormat': response.getheader('Content-Type')
                            })
                        if not self.fetch_remote:
                            self._jsonld['sdDatePublished'] = iso_now()
                    if self.fetch_remote:
                        out_file_path.parent.mkdir(parents=True, exist_ok=True)
                        urllib.request.urlretrieve(response.url, out_file_path)
        elif self.source is None:
            # Allows to record a File entity whose @id does not exist, see #73
            warnings.warn(f"No source for {self.id}")
        else:
            out_file_path.parent.mkdir(parents=True, exist_ok=True)
            if not out_file_path.exists() or not out_file_path.samefile(self.source):
                shutil.copy(self.source, out_file_path)



#!/usr/bin/env python

# Copyright 2019-2023 The University of Manchester, UK
# Copyright 2020-2023 Vlaams Instituut voor Biotechnologie (VIB), BE
# Copyright 2020-2023 Barcelona Supercomputing Center (BSC), ES
# Copyright 2020-2023 Center for Advanced Studies, Research and Development in Sardinia (CRS4), IT
# Copyright 2022-2023 École Polytechnique Fédérale de Lausanne, CH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from collections.abc import MutableMapping

from dateutil.parser import isoparse
from .. import vocabs


class Entity(MutableMapping):

    def __init__(self, crate, identifier=None, properties=None):
        self.crate = crate
        if identifier:
            self.id = self.format_id(identifier)
        else:
            self.id = f"#{uuid.uuid4()}"
        if properties:
            empty = self._empty()
            empty.update(properties)
            self._jsonld = empty
        else:
            self._jsonld = self._empty()

    # Format the given ID with rules appropriate for this type.
    # For example, Dataset (directory) data entities SHOULD end with /
    def format_id(self, identifier):
        return str(identifier)

    def __repr__(self):
        return f"<{self.id} {self.type}>"

    def properties(self):
        return self._jsonld

    def as_jsonld(self):
        return self._jsonld

    @property
    def _default_type(self):
        clsName = self.__class__.__name__
        if clsName in vocabs.RO_CRATE["@context"]:
            return clsName
        return "Thing"

    def canonical_id(self):
        return self.crate.resolve_id(self.id)

    def __hash__(self):
        return hash(self.canonical_id())

    def _empty(self):
        val = {
            "@id": self.id,
            "@type": self._default_type
        }
        return val

    def __getitem__(self, key):
        v = self._jsonld[key]
        if v is None or key.startswith("@"):
            return v
        values = v if isinstance(v, list) else [v]
        deref_values = []
        for entry in values:
            if isinstance(entry, dict):
                try:
                    id_ = entry["@id"]
                except KeyError:
                    raise ValueError(f"no @id in {entry}")
                else:
                    deref_values.append(self.crate.get(id_, id_))
            else:
                deref_values.append(entry)
        return deref_values if isinstance(v, list) else deref_values[0]

    def __setitem__(self, key: str, value):
        if key.startswith("@"):
            raise KeyError(f"cannot set '{key}'")
        values = value if isinstance(value, list) else [value]
        ref_values = [{"@id": _.id} if isinstance(_, Entity) else _ for _ in values]
        self._jsonld[key] = ref_values if isinstance(value, list) else ref_values[0]

    def __delitem__(self, key: str):
        if key.startswith("@"):
            raise KeyError(f"cannot delete '{key}'")
        del self._jsonld[key]

    def popitem(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError

    def __iter__(self):
        return iter(self._jsonld)

    def __len__(self):
        return len(self._jsonld)

    def __contains__(self, key):
        return key in self._jsonld

    def __eq__(self, other):
        if not isinstance(other, Entity):
            return NotImplemented
        return self.id == other.id and self._jsonld == other._jsonld

    @property
    def type(self):
        return self._jsonld['@type']

    @property
    def datePublished(self):
        d = self.get('datePublished')
        return d if not d else isoparse(d)

    @datePublished.setter
    def datePublished(self, value):
        try:
            value = value.isoformat()
        except AttributeError:
            pass
        self['datePublished'] = value

    def delete(self):
        self.crate.delete(self)

    def append_to(self, key: str, value, compact=False):
        if key.startswith("@"):
            raise KeyError(f"cannot append to '{key}'")
        current_value = self._jsonld.setdefault(key, [])
        if not isinstance(current_value, list):
            current_value = self._jsonld[key] = [current_value]
        if not isinstance(value, list):
            value = [value]
        current_value.extend([{"@id": _.id} if isinstance(_, Entity) else _ for _ in value])
        if compact and len(current_value) == 1:
            self._jsonld[key] = current_value[0]
